// TrackerChecker.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

enum TRACKER_STATE
{
	TRACKER_STATE_UNKNOWN,
	TRACKER_STATE_RESOLVE_FAILED,
	TRACKER_STATE_RESOLVED,
	TRACKER_STATE_CONNECTING,
	TRACKER_STATE_CONNECTED,
	TRACKER_STATE_CONNECT_ERROR,
	TRACER_STATE_DISCONNECTED,
	TRACKER_STATE_OTHER_ERROR
};

struct TrackerInfo
{
	TrackerInfo() : m_state(TRACKER_STATE_UNKNOWN), m_s(INVALID_SOCKET) {}
	TrackerInfo(const string& str) : m_state(TRACKER_STATE_UNKNOWN), m_addrStr(str), m_s(INVALID_SOCKET){}

	string m_addrStr;
	sockaddr_in m_addr;
	enum TRACKER_STATE m_state;
	SOCKET m_s;
};


string GetIniString(const string& section, const string& key, const string& def, const string& ini)
{
	char buf[4096];
	GetPrivateProfileString(section.c_str(), key.c_str(), def.c_str(), buf, sizeof(buf), ini.c_str());
	return string(buf);
}

////////////////////////////////////////////////////////////////////////////////
string itoa(int num)
{
	char buf[64];
	return string(itoa(num, buf, 10));
}


typedef vector<TrackerInfo> VTI;
typedef vector<size_t> VI;

int main(int argc, _TCHAR* argv[])
{
	int cp = GetConsoleCP();
	cp = GetACP();
	SetConsoleCP(GetACP());
	SetConsoleOutputCP(GetACP());
	WSADATA wsaData;
	int res = WSAStartup(MAKEWORD( 2, 2 ), &wsaData); //TODO: check error


	char progPath[MAX_PATH];
	res = GetModuleFileName(NULL, progPath, sizeof(progPath));
	string programDir;
	{
		char fullPath[MAX_PATH];
		char *filePart;
		res = GetFullPathName(progPath, sizeof(fullPath), fullPath, &filePart);
		programDir.assign(fullPath, filePart);
	}
	string mainIni;
	{
		char mainIniPath[MAX_PATH];
		char *mainInifilePart;
		res = GetFullPathName(progPath, sizeof(mainIniPath), mainIniPath, &mainInifilePart);
		strcpy(mainInifilePart, "Settings.ini");
		mainIni = mainIniPath;
	}
	string trackerIniIdx(GetIniString("Settings", "TrackerIniIndex", "", mainIni));
	int city = -1;
	int provider = -1;
	sscanf(trackerIniIdx.c_str(), "%d %d", &city, &provider);
	city++;
	provider++;
	const string trackerIni(GetIniString("Settings", "TrackersFile", "", mainIni));
	string cityStr(GetIniString("", itoa(city), "", trackerIni));
	if (cityStr.size() == 0)
	{
		printf("Bad city %d\n", city);
		return 1;
	}
	string providerStr(GetIniString((string(" ") + cityStr), itoa(provider), "", trackerIni));
	if (providerStr.size() == 0)
	{
		printf("Bad provider %d for city %s\n", provider, cityStr.c_str());
		return 1;
	}
	string providerSection((string(" ") + cityStr + " " + providerStr));
	int trackerNum = GetPrivateProfileInt(providerSection.c_str(), "", 0, trackerIni.c_str());
	if (trackerNum == 0)
	{
		printf("No active trackers\n");
		return 1;
	}
	printf("Checking  %s, %d providers\n", providerSection.c_str(), trackerNum);

	VTI vti;
	for(int i = 1; i <= trackerNum; ++i)
	{
		string tracker(GetIniString(providerSection, itoa(i).c_str(), "", trackerIni));
		if (tracker.size() > 0)
		{
			vti.push_back(TrackerInfo(tracker));
		}
	}
	cp = GetConsoleCP();
	cp = GetACP();
/*
	for(int k = 0; k < 10; ++k)
	{
		vti.push_back(TrackerInfo("http://nest:2170"));
		vti.push_back(TrackerInfo("http://127.0.0.1:2170/announce"));
		vti.push_back(TrackerInfo("127.0.0.1:2170/announce"));
		vti.push_back(TrackerInfo("http://127.0.0.1/announce"));
		vti.push_back(TrackerInfo("http://172.17.243.50:2170"));
		vti.push_back(TrackerInfo("http://nest:2170/announce"));
		vti.push_back(TrackerInfo("http://nest/announce"));
		vti.push_back(TrackerInfo("http://nest"));
		vti.push_back(TrackerInfo("http://172.17.243.50"));
		vti.push_back(TrackerInfo("nest"));
		vti.push_back(TrackerInfo("127.0.0.1"));
		vti.push_back(TrackerInfo("user.rtsnet.ru"));
		vti.push_back(TrackerInfo("172.17.242.87"));
		vti.push_back(TrackerInfo("ya.ru"));
		vti.push_back(TrackerInfo("ya.ru"));
	}
*/
#if 1
	VTI::iterator iter = vti.begin();
	for( ; iter != vti.end(); ++iter)
	{
		TrackerInfo& ti(*iter);
		// cut off prefix
		const string prefix("http://");
		string str;
		if ((ti.m_addrStr.size() >= prefix.size()) && (ti.m_addrStr.substr(0, prefix.size()) == prefix))
		{
			str = ti.m_addrStr.substr(prefix.size());
		}
		else
		{
			str = ti.m_addrStr;
		}
		string address;
		int port;
		const int default_port = 80;
		//      :, /   
		string::size_type pos = str.find_first_of(":/");
		if (pos == string::npos)
		{
			address = str;
			port = default_port;
		}
		else
		{
			address = str.substr(0, pos);
			if (str[pos] == ':')
			{
				port = atoi(str.substr(pos + 1).c_str());
			}
			else
			{
				port = default_port;
			}
		}
		//printf("str %s => address %s, port %d\n", str.c_str(), address.c_str(), port);
		char buf[20];
		itoa(port, buf, 10);
		struct addrinfo *pai = NULL;
		struct addrinfo aiHints;
		memset(&aiHints, 0, sizeof(aiHints));
		aiHints.ai_family = AF_INET;
		aiHints.ai_socktype = SOCK_STREAM;
		aiHints.ai_protocol = IPPROTO_TCP;
		res = getaddrinfo(address.c_str(), buf, &aiHints, &pai);
		if ((res != 0) || (pai == NULL))
		{
			ti.m_state = TRACKER_STATE_RESOLVE_FAILED;
		}
		else
		{
			ti.m_addr = *(sockaddr_in*)(pai->ai_addr);
			freeaddrinfo(pai);
		}
		ti.m_s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (ti.m_s == INVALID_SOCKET)
		{
			ti.m_state = TRACKER_STATE_OTHER_ERROR;
		}
		else
		{
			u_long param = 1;
			ioctlsocket(ti.m_s, FIONBIO, &param); //TODO: error
			res = connect(ti.m_s, (sockaddr *)&ti.m_addr, sizeof(ti.m_addr));
			if (res == 0)
			{
				ti.m_state = TRACKER_STATE_CONNECTED;
			}else
			{
				if (WSAGetLastError() == WSAEWOULDBLOCK)
				{
					ti.m_state = TRACKER_STATE_CONNECTING;
				}
				else
				{
					ti.m_state = TRACKER_STATE_CONNECT_ERROR;
				}
			}
			
		}
	}
	DWORD t0 = GetTickCount();
	while(true)
	{
		size_t rem = 0;
		for(size_t j = 0 ; (j < vti.size()); ++j)
		{
			if (vti[j].m_state == TRACKER_STATE_CONNECTING)
			{
				rem++;
			}
		}
		printf("<-----%d remains %d\n", GetTickCount() - t0, rem);
		timeval tv = {1, 0};
		fd_set read, write, except;
		FD_ZERO(&read);
		FD_ZERO(&write);
		FD_ZERO(&except);
		iter = vti.begin();
		size_t ss = vti.size();
		const int maxConnections = 10; //FD_SETSIZE
		int scnt = 0;
		VI vi;
		for(size_t j = 0 ; (j < vti.size()) && (scnt < FD_SETSIZE); ++j)
		{
			if (vti[j].m_state == TRACKER_STATE_CONNECTING)
			{
				FD_SET(vti[j].m_s, &write);
				FD_SET(vti[j].m_s, &read);
				FD_SET(vti[j].m_s, &except);
				scnt++;
				vi.push_back(j);
			}
		}
		if (scnt >= maxConnections)
		{
			printf("Out of set size\n");
		}
		if (scnt == 0)
		{
			break;
		}
		int cnt = select(0, &read, &write, &except, &tv);
		for(int i = 0; i < scnt; ++i)
		{
			TrackerInfo& ti = vti[vi[i]];
			if (FD_ISSET(vti[vi[i]].m_s, &write))
			{
				printf("%s connected ok\n", vti[vi[i]].m_addrStr.c_str());
				vti[vi[i]].m_state = TRACKER_STATE_CONNECTED;
				continue;
			}
			if (FD_ISSET(vti[vi[i]].m_s, &read))
			{
				printf("%d read\n", vi[i]);
			}
			if (FD_ISSET(vti[vi[i]].m_s, &except))
			{
				printf("%s can't connect\n", vti[vi[i]].m_addrStr.c_str());
				vti[vi[i]].m_state = TRACKER_STATE_CONNECT_ERROR;
				continue;
			}
		}
		//TODO: select
		//TODO: check state
		printf(">-----%d\n", GetTickCount() - t0);
	}
	DWORD t1 = GetTickCount();
	printf("Total %d msec\n", t1 - t0);
	iter = vti.begin();
	for( ; iter != vti.end(); ++iter)
	{
		//TODO: clean connecting
	}
#endif
	string ok;
	for(size_t j = 0 ; (j < vti.size()); ++j)
	{
		if (vti[j].m_state == TRACKER_STATE_CONNECTED)
		{
			if (ok.size() != 0)
			{
				ok += " ";
			}
			ok += itoa(j + 1);
		}
	}
	printf("TrackerCheck=%s 0\n", ok.c_str());
	WritePrivateProfileString("Settings", "TrackerCheck", (ok + " 0").c_str(), mainIni.c_str());

	res = WSACleanup();
	return 0;
}
