// bencode.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "bencode.h"

string g_mainIni;
////////////////////////////////////////////////////////////////////////////////
typedef vector<int> VI;
////////////////////////////////////////////////////////////////////////////////
FILE *fLog = NULL;
////////////////////////////////////////////////////////////////////////////////
void Log(const char *str)
{
	time_t t;
	time(&t);
	char *p = asctime(localtime(&t));
	p[strlen(p)-1] = ' ';
	p[strlen(p)] = 0;
	fputs(p, fLog);
	fputs(str, fLog);
	fputs("\n", fLog);
}
////////////////////////////////////////////////////////////////////////////////
string GetIniString(const string& section, const string& key, const string& def, const string& ini)
{
	char buf[4096];
	GetPrivateProfileString(section.c_str(), key.c_str(), def.c_str(), buf, sizeof(buf), ini.c_str());
	return string(buf);
}
////////////////////////////////////////////////////////////////////////////////
string itoa(int num)
{
	char buf[64];
	return string(itoa(num, buf, 10));
}
////////////////////////////////////////////////////////////////////////////////
string urlencode(const string url)
{
	string result;	
	for(size_t pos = 0; pos < url.size(); ++pos)
	{
		char ch = url[pos];
		if ((ch >= 'a') && (ch <= 'z'))
		{
			result += ch;
		}
		else if ((ch >= 'A') && (ch <= 'Z'))
		{
			result += ch;
		}
		else if ((ch >= '0') && (ch <= '9'))
		{
			result += ch;
		}
		else if (ch == '.')
		{
			result += ch;
		}
		else
		{
			char buf[10];
			unsigned char chl = ch;
			sprintf(buf, "%%%02x", chl);
			result += buf;

		}
	}
	return result;
}
////////////////////////////////////////////////////////////////////////////////
static be_node *createListOfString(const string& str)
{
	be_node *p = new be_node(BE_STR);
	p->setStr(str);
	be_node *pl = new be_node(BE_LIST);
	if (str != "")
	{
		pl->pushToList(p);
	}
	return pl;
}

////////////////////////////////////////////////////////////////////////////////
void RunUTorrent(const string& program, const string& progArgs, const string& torrentFileName)
{
	STARTUPINFO si;
	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = 0;
	PROCESS_INFORMATION pi;

	size_t pos = progArgs.find("%1");
	string args;
	if (pos == string::npos)
	{
		Log((string("Wrong launch arguments, no way to add filename to \"") + progArgs + string("\"")).c_str());
		args = progArgs;
	}
	else
	{
		args = progArgs.substr(0, pos) + "\"" + torrentFileName + "\"" + progArgs.substr(pos + 2);
	}
	string cmdLine(string("\"") +  program + "\" " + args);
	char buf[1024];
	strcpy(buf, cmdLine.c_str());
	BOOL bRes = CreateProcess(NULL, buf, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi);
	if (bRes)
	{
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
	else
	{
		Log((string("Running of \"") + buf + "\" failed with code " + itoa(GetLastError())).c_str());
	}
}
////////////////////////////////////////////////////////////////////////////////
bool ReadCompleteFile(const char *fileName, char *&p, size_t &size)
{
	FILE *f = fopen(fileName, "rb");
	if (f == NULL)
	{
		return false;
	}
	fseek(f, 0, SEEK_END);
	size = ftell(f);
	fseek(f, 0, SEEK_SET);
	if (f == 0)
	{
		fclose(f);
		return false;
	}
	p = new char[size];
	size_t readSize = fread(p, 1, size, f);
	fclose(f);
	if (readSize != size)
	{
		delete [] p;
		return false;
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
void GetWorkingTrackers(const char *pTrackerStr, VI& trackers)
{
	const char *pbegin = pTrackerStr;
	const char *pend = NULL;
	while(true)
	{
		int li = strtol( pbegin, const_cast<char **>(&pend), 10 );
		if (li == 0)
		{
			break;
		}
		trackers.push_back(li);
		
		pbegin = pend;
		while(*pbegin == ' ') pbegin++;
		if (*pbegin == 0) break;
	}
}
////////////////////////////////////////////////////////////////////////////////
BEINT64 CalcTorrentSize(be_node& root)
{
	be_node& info = *(root["info"]);
	if (info["length"] != NULL)
	{	//   ,  length    info
		return info["length"]->getInt(); 
	}
	//     
	BEINT64 torrentSize = 0;
	bv& fileList = info["files"]->getList_();
	for(size_t i = 0; i < fileList.size(); ++i)
	{
		be_node& file = *(fileList[i]);
		torrentSize += file["length"]->getInt();
	}
	return torrentSize;
}
////////////////////////////////////////////////////////////////////////////////
int AddTrackerList(int city, int provider, be_node& root, VI& trackers)
{ // enum all trackers
	const string trackerIni(GetIniString("Settings", "TrackersFile", "", g_mainIni));
	const bool addStat = stricmp(GetIniString("Settings", "AddStat", "", g_mainIni).c_str(), "True") == 0;
	const string cityStr(GetIniString("", itoa(city), "", trackerIni));
	if (cityStr.size() == 0)
	{
		char buf[256];
		sprintf(buf, "Bad city %d", city);
		Log(buf);
		return 1;
	}
	string providerStr(GetIniString((string(" ") + cityStr), itoa(provider), "", trackerIni));
	if (providerStr.size() == 0)
	{
		char buf[256];
		sprintf(buf, "Bad provider %d for city %s", provider, cityStr.c_str());
		Log(buf);
		return 1;
	}
	string providerSection((string(" ") + cityStr + " " + providerStr));
	int trackerNum = GetPrivateProfileInt(providerSection.c_str(), "", 0, trackerIni.c_str());
	if (trackerNum == 0)
	{
		char buf[256];
		sprintf(buf, "No active retrackers for provider %s, city %s", providerStr.c_str(), cityStr.c_str());
		Log(buf);
		return 1;
	}
	be_node *pAnn = root["announce"];
	const string announce_ = (pAnn == NULL) ? "" : pAnn->getStr();
	const string comment = (root["comment"] != NULL) ? root["comment"]->getStr() : "";
	const string name = (*root["info"])["name"]->getStr();

	be_node *pOuterList = root["announce-list"];
	if (pOuterList == NULL)
	{
		pOuterList = new be_node(BE_LIST);
		root.addToDict("announce-list", pOuterList);
	}
	pOuterList->pushToList(createListOfString(announce_));
	char torrentSizeStr[256];
	sprintf(torrentSizeStr, "%I64d", CalcTorrentSize(root));
	for(size_t i = 0; i < trackers.size(); i++)
	{
		if (trackers[i] > trackerNum)
		{
			continue;
		}
		string url(GetIniString(providerSection, itoa(trackers[i]), "", trackerIni));
		if (url.size() == 0)
		{
			continue;
		}
		if ((url.find("re-tracker.ru") != string::npos) || addStat)
		{
			// ?name=<torrentName>&size=<TorrentSize>&comment=<Comment>&isp=<city>+<provider>
			url += string("?name=") + urlencode(name) + string("&size=") + string(torrentSizeStr);
			url += "&comment=" + urlencode(comment) + "&isp=" + itoa(city) + "+" + itoa(provider);
		}
		pOuterList->pushToList(createListOfString(url));
	}
	return 0;
}
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
	if (argc < 2)
	{
		return 0;
	}
	//DebugBreak();
	bool justPatch = (argc > 2);
	try{
	char progPath[MAX_PATH];
	DWORD res = GetModuleFileName(NULL, progPath, sizeof(progPath));
	string programDir;
	{
		char fullPath[MAX_PATH];
		char *filePart;
		res = GetFullPathName(progPath, sizeof(fullPath), fullPath, &filePart);
		programDir.assign(fullPath, filePart);
	}
	fLog = fopen((programDir + "TorrentAdder.log").c_str(), "aw");
	if (fLog == NULL)
	{
		return 1;
	}
	{
		char buf[256];
		sprintf(buf, "Running for torrent \"%s\"", argv[1]);
		Log(buf);
	}
/*
	char iniPath[MAX_PATH];
	char *filePart;
	res = GetFullPathName(progPath, sizeof(iniPath), iniPath, &filePart);
	strcpy(filePart, "trackerssimple.ini");
*/
	char *origTorrentName;
	char fullOrigFile[1024];
	res = GetFullPathName(argv[1], sizeof(fullOrigFile), fullOrigFile, &origTorrentName);
	{
		char mainIniPath[MAX_PATH];
		char *mainInifilePart;
		res = GetFullPathName(progPath, sizeof(mainIniPath), mainIniPath, &mainInifilePart);
		strcpy(mainInifilePart, "Settings.ini");
		g_mainIni = mainIniPath;
	}
	char *p;
	size_t size;
	if (!ReadCompleteFile(argv[1], p, size))
	{
		return 1;
	}
	auto_ptr<be_node>  bn_(be_node::decode(p, size));

	if (bn_.get() == NULL)
	{
		Log("Bad torrent");
		return 1;
	}
	be_node& root = *bn_;

	//    
	int city = -1;
	int provider = -1;
	{
		string trackerIniIdx(GetIniString("Settings", "TrackerIniIndex", "", g_mainIni));
		sscanf(trackerIniIdx.c_str(), "%d %d", &city, &provider);
		if ((city == -1) || (provider == -1))
	{
			Log("Bad city or provider");
			return 1;
	}
		city++;
		provider++;
	}
	//   
	VI trackers;
	{
		string trackerCheck(GetIniString("Settings", "TrackerCheck", "", g_mainIni));
		GetWorkingTrackers(trackerCheck.c_str(), trackers);
	}


	
	int ret = AddTrackerList(city, provider, root, trackers);
	if (ret != 0)
	{
		return ret;
	}
	//bn->dump();

	//TODO: calc better
	char *p1 = (char *)malloc(size*10);
	char *p2 = p1;

	//   -
	string ourTorrent(programDir + origTorrentName);
	if (justPatch)
	{
		ourTorrent = argv[2];
	}
	{
		FILE *fo = fopen(ourTorrent.c_str(), "wb");
		root.save(p2, size);
		size_t size2 = fwrite(p1, 1, p2 - p1, fo);
		fclose(fo);
	}
	//run utorrent witn new .torrent
	if (!justPatch)
	{
		string program(GetIniString("Settings", "LaunchPath", "", g_mainIni));
		string progArgs(GetIniString("Settings", "LaunchArguments", "", g_mainIni));
		RunUTorrent(program, progArgs, ourTorrent);
	}
	delete []p;
	}catch(BadType&)
	{
		Log("Bad type access");
		fclose(fLog);
		return 1;
	}
	catch(BadInput& )
	{
		Log("Bad input data");
		fclose(fLog);
		return 1;
	}
	catch(...)
	{
		Log("Shit happens");
		fclose(fLog);
		return 1;
	}
	fclose(fLog);
	return 0;
}
////////////////////////////////////////////////////////////////////////////////
int WINAPI WinMain(HINSTANCE instance, HINSTANCE prev_instance, char* command_line, int show_command)
{
    return main(__argc, __argv);
}
