#ifndef _BENCODE_H
#define _BENCODE_H

#include <stddef.h>
typedef ptrdiff_t ssize_t;

typedef __int64 BEINT64;

////////////////////////////////////////////////////////////////////////////////
//TODO: make proper exceptions
class BadType
{
public:
	string m_reason;
};
////////////////////////////////////////////////////////////////////////////////
class BadInput
{
public:
	string m_reason;
};
////////////////////////////////////////////////////////////////////////////////
class NoKeyInDict
{
public:
	string m_reason;
};

enum BE_TYPE{
	BE_STR,
	BE_INT,
	BE_LIST,
	BE_DICT,
};

class be_node;


class bv : public vector<be_node *>
{
public:
	~bv();
};

class dict : public map<string, be_node *>
{
public:
	~dict();
};


class be_node {
public:
	enum BE_TYPE m_type;
public:
	be_node(enum BE_TYPE type) : m_type(type){}
	be_node(const string& str) : m_type(BE_STR), m_str(str){}

	bool save(char *&benc, size_t size);
	void dump()const;

	BEINT64 getInt(void)const
	{
		if (m_type == BE_INT)
		{
			return m_int;
		}
		throw BadType();
	}

	string getStr(void)const
	{
		if (m_type == BE_STR)
		{
			return m_str;
		}
		throw BadType();
	}

	void setStr(const string& str)
	{
		if (m_type == BE_STR)
		{
			m_str = str;
			return;
		}
		throw BadType();
	}

	bv& getList_(void)
	{
		if (m_type == BE_LIST)
		{
			return m_list;
		}
		throw BadType();
	}

	void pushToList(be_node *bn)
	{
		if (m_type != BE_LIST)
		{
			throw BadType();
		}
		m_list.push_back(bn);
	}

/*
	dict& getDict_(void)
	{
		if (m_type == BE_DICT)
		{
			return m_dict;
		}
		throw BadType();
	}
*/
	void addToDict(const char *key, be_node *bn)
	{
		if (m_type != BE_DICT)
		{
			throw BadType();
		}
		m_dict[key] = bn;
	}


	be_node *operator[](const char *key)
	{
		if (m_type != BE_DICT)
		{
			throw BadType();
		}
		dict::iterator iter = m_dict.find(key);
		if (iter == m_dict.end())
		{
			return NULL;
		}
		else
		{
			return iter->second;
		}
	}


private:
	void _dump(ssize_t indent)const;

private:
	BEINT64 m_int;
	string m_str;
	bv m_list;
	dict m_dict;

public:
	static be_node *decode(const char *bencode, size_t size);
private:
	static be_node *intDecode(const char **data, size_t& size);
};



#endif
