#include "stdafx.h"
#include "bencode.h"

static BEINT64 parseInt(const char **data, size_t& size)
{
	char *endp;
	BEINT64 ret = _strtoi64(*data, &endp, 10);
	size -= endp - *data;
	*data = endp;
	return ret;
}

static void parseStr(const char **data, size_t& size, string& s)
{
	size_t len = static_cast<size_t>(parseInt(data, size));
	if (**data == ':') {
		s.assign(*data + 1, len);
		*data += len + 1;
		size -= len + 1;
	}
	else
	{
		throw BadInput();
	}
}

be_node *be_node::intDecode(const char **data, size_t& size)
{
	be_node *ret = NULL;

	if (size < 1)
	{
		return NULL;
	}

	switch (**data) {
		/* lists */
		case 'l': {

			ret = new be_node(BE_LIST);

			++(*data);
			--size;
			while (**data != 'e') {
				ret->m_list.push_back(intDecode(data, size));
			}
			++(*data);
			--size;

			return ret;
		}

		/* dictionaries */
		case 'd': {
			unsigned int i = 0;

			ret = new be_node(BE_DICT);

			++(*data);
			--size;
			while (**data != 'e') {
				string key;
				parseStr(data, size, key);
				ret->m_dict[key] = intDecode(data, size);
			}
			++(*data);
			--size;

			return ret;
		}

		/* integers */
		case 'i': {
			ret = new be_node(BE_INT);

			++(*data);
			--size;
			ret->m_int = parseInt(data, size);
			if (**data != 'e')
				return NULL;
			++(*data);
			--size;

			return ret;
		}

		/* byte strings */
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		{
			string s;
			parseStr(data, size, s);
			return new be_node(s);
		}

		/* invalid */
		default:
			break;
	}

	return ret;
}

be_node *be_node::decode(const char *data, size_t size)
{
	size_t size_local = size;
	be_node *node =intDecode(&data, size_local);
	if (size_local != 0)
	{
		printf("Bad torrent: size %d, tail %d\n", size, size_local);
	}
	return node;
}


//TODO: why ssize_t
static void _be_dump_indent(ssize_t indent)
{
	while (indent-- > 0)
		printf("    ");
}
void be_node::_dump(ssize_t indent)const
{
	size_t i;

	_be_dump_indent(indent);
	indent = (indent < 0) ? -indent : indent;

	switch (m_type) {
		case BE_STR:
			printf("str = %s\n", m_str.c_str());
			break;

		case BE_INT:
			printf("int = %I64di\n", m_int);
			break;

		case BE_LIST:
			puts("list [");

			for (i = 0; i < m_list.size(); ++i)
				m_list[i]->_dump(indent + 1);

			_be_dump_indent(indent);
			puts("]");
			break;

		case BE_DICT:
			puts("dict {");

			dict::const_iterator iter = m_dict.begin();
			for( ; iter != m_dict.end(); ++iter){
				_be_dump_indent(indent + 1);
				printf("%s => ", iter->first.c_str());
				iter->second->_dump(-(indent + 1));
			}

			_be_dump_indent(indent);
			puts("}");
			break;
	}
}
void be_node::dump(void)const
{
	_dump(0);
}
bool be_node::save(char *&benc, size_t size)
{
	char *benc_orig = benc;
	size_t i;

	switch (m_type) {
		case BE_STR:
			sprintf(benc, "%d:", m_str.size());
			benc += strlen(benc);
			memcpy(benc, m_str.data(), m_str.size());
			benc += m_str.size();
			break;

		case BE_INT:
			sprintf(benc, "i%I64de", m_int);
			benc += strlen(benc);
			break;

		case BE_LIST:
			*benc++ = 'l';

			for (i = 0; i < m_list.size(); ++i)
			{
				m_list[i]->save(benc, size);
			}
			*benc++ = 'e';
			break;

		case BE_DICT:
			*benc++ = 'd';

			dict::const_iterator iter = m_dict.begin();
			for( ; iter != m_dict.end(); ++iter){
				sprintf(benc, "%d:", iter->first.size());
				benc += strlen(benc);
				memcpy(benc, iter->first.data(), iter->first.size());
				benc += iter->first.size();
				iter->second->save(benc, size);
			}
			*benc++ = 'e';
			break;
	}
	return 0;
}

bv::~bv()
{
	iterator it = begin();
	for( ; it != end(); ++it)
	{
		delete *it;
	}
}

dict::~dict()
{
	iterator it = begin();
	for( ; it != end(); ++it)
	{
		delete it->second;
	}
}
